/*
 * makeztxt: Translates ASCII files into zTXT databases for the Palm Handheld
 * Copyright (C) 2001 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <getopt.h>
#include <regex.h>

#include <zlib.h>
#include "databaseHdrs.h"
#include <gutenpalm_common.h>


/*  makeztxt program version */
#define VERSION                 "1.10"
/*  Current version of the zTXT format (v1.10) */
#define ZTXT_VERSION            0x010A
/*  Palm's have an epoch of Jan. 1, 1904 */
#define PALM_CTIME_OFFSET       0x7C25B080
/*  Max size of the header.  Should be fine except for huge docs */
#define MAX_HEADER_SIZE         4096
/*  makeztxt options file in user home directory */
#define OPTIONS_FILE            ".gutenpalmrc"



/* A linked list of regular expressions */
typedef struct regex_nodeType {
  char          *pattern;
  regex_t       *prog;
  int           bad;
  struct regex_nodeType *next;
} regex_node;

/* A simple struct for bookmarks generated from the regex */
typedef struct bmrk_nodeType {
  UInt32        offset;
  char          title[MAX_BMRK_LENGTH];
  struct bmrk_nodeType *next;
} bmrk_node;

/* User definable options for new databases */
typedef struct optsType {
  short         version;
  char          *creator;
  char          *type;
  char          *title;
  int           wbits;
  int           list_bmrks;
  int           num_regex;
  short         num_bookmarks;
  regex_node    *regex_list;
  bmrk_node     *bookmarks;
  int           adjust_type;
  int           line_length;
} opts;



/* Process commandline arguments  */
int     process_arguments(int, char **);

/*  Load text and adjust linebreaks for the Palm  */
u_int   load_and_adjust(FILE *, char *);
char *  fgets_nocr(char *, int, FILE *);

/*  Do regular expression searching for bookmarks  */
void    process_regex(char *);
void    add_regex(char *);
void    add_bmrk(char *, u_long, u_long);
void    clear_regex_bmrks(void);
u_int   output_bookmarks(int);

/*  Read option file in home directory for regex  */
FILE *  fopen_in_home_dir(char *, char *);
void    load_options(void);

/*  Converts a 4 byte string Palm ID into a 4 byte integer  */
UInt32  id_to_int(char *);

/*  Fatal error if condition is true  */
void    errorif(int);


/* Global options */
opts    options;



int
main(int argc, char *argv[])
{
  int           last_arg;
  char          temptitle[dmDBNameLength];
  int           x;
  char          *y;
  int           zcode;
  int           data_offset;
  int           curr_record;
  int           num_records;
  int           num_data_records;
  char          *inputbuf;
  char          *zbuf;
  char          *header;
  char          outfilename[128];
  u_int         zbuf_size;
  z_stream      zstream;

  zTXT_record0  record0;
  DatabaseHdrType *dbHeader;
  RecordEntryType *dbRecordEntries;
  FILE          *infile;
  int           outfile;
  struct stat   instats;
  u_int         filesize;
  bmrk_node     *current;
  char          tmpstr[MAX_BMRK_LENGTH + 1];


  /* Default options for a new DB */
  options.version       = ZTXT_VERSION;
  options.creator       = GPLM_CREATOR_ID;
  options.type          = ZTXT_TYPE_ID;
  options.title         = NULL;
  options.wbits         = MAXWBITS;
  options.num_regex     = 0;
  options.num_bookmarks = 0;
  options.list_bmrks    = 0;
  options.regex_list    = NULL;
  options.bookmarks     = NULL;
  options.adjust_type   = 0;
  options.line_length   = 0;

  last_arg = process_arguments(argc, argv);
  load_options();

  if (last_arg >= argc)
    {
      fprintf(stderr, "no input\n");
      exit(1);
    }

  if (options.title == NULL)
    { 
      strncpy(temptitle, argv[last_arg], dmDBNameLength);
      options.title = temptitle;
    }

  /* Open the input file */
  infile = fopen(argv[last_arg], "r");
  errorif(infile == NULL);

  x = stat(argv[last_arg], &instats);
  errorif(x == -1);

  /* Allocate the input buffer */
  inputbuf = (char *)malloc(instats.st_size + 1);
  errorif(inputbuf == NULL);

  /* Load and process the input file */
  filesize = load_and_adjust(infile, inputbuf);
  errorif(filesize == 0);
  fclose(infile);

  /* Process the regular expressions to generate some bookmarks */
  if (options.num_regex > 0)
      process_regex(inputbuf);

  if (options.list_bmrks)
    {
      printf("Generated bookmarks\nOffset\t\tTitle\n"
             "-----------\t--------------------\n");
      current = options.bookmarks;
      while (current != NULL)
        {
          strncpy(tmpstr, current->title, MAX_BMRK_LENGTH);
          tmpstr[MAX_BMRK_LENGTH] = '\0';
          printf("%ld\t\t%s\n",current->offset, tmpstr);
          current = current->next;
        }
    }

  /* Allocate the compression buffer.  Large than the input to be safe */
  zbuf_size = instats.st_size + (.02 * instats.st_size);
  zbuf = (char *)malloc(zbuf_size);
  errorif(zbuf == NULL);

  /* Allocate a chunk of memory to hold the DB header */
  header = (char *)malloc(MAX_HEADER_SIZE);
  errorif(header == NULL);

  /* Initialize the Palm DB header structure */
  dbHeader = (DatabaseHdrType *)header;
  strncpy(dbHeader->name, options.title, dmDBNameLength);
  dbHeader->attributes = htonl(0x0008);
  dbHeader->version = htons(options.version);
  dbHeader->creationDate = htonl(time(NULL) + PALM_CTIME_OFFSET);
  dbHeader->modificationDate = htonl(time(NULL) + PALM_CTIME_OFFSET);
  dbHeader->lastBackupDate = 0;
  dbHeader->modificationNumber = 0;
  dbHeader->appInfoID = 0;
  dbHeader->sortInfoID = 0;
  dbHeader->type = htonl(id_to_int(options.type));
  dbHeader->creator = htonl(id_to_int(options.creator));
  dbHeader->uniqueIDSeed = 0;
  dbHeader->recordList.nextRecordListID = 0;
  dbRecordEntries = (RecordEntryType *)&(dbHeader->recordList.firstEntry);




  /* These values must be set before using Zlib */
  zstream.zalloc = (alloc_func)0;
  zstream.zfree = (free_func)0;
  zstream.opaque = (voidpf)0;
  zstream.next_in = inputbuf;
  zstream.next_out = zbuf;
  zstream.avail_in = instats.st_size;
  zstream.avail_out = zbuf_size;

  /* Initialize the compression stream */
  x = deflateInit2(&zstream, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
                   options.wbits, 9, Z_DEFAULT_STRATEGY);
  errorif(x != Z_OK);

  /* Compress the input file into one big buffer */
  zcode = deflate(&zstream, Z_FINISH);
  if (zcode != Z_STREAM_END)
    {
      fprintf(stderr,
              "Input still pending...\n"
              "\t avail_in = %d\n"
              "\tavail_out = %d\n",zstream.avail_in,zstream.avail_out);
      exit(1);
    }
  deflateEnd(&zstream);




  /* Calculate the offset in the DB where the data will go */
  data_offset = sizeof(DatabaseHdrType) - sizeof(UInt16);
  num_records = zstream.total_out / RECORD_SIZE;
  if ((num_records * RECORD_SIZE) != zstream.total_out)
    num_records++;
  /* One extra for record #0 */
  num_records++;
  /* Add a record for the bookmarks if there are any */
  if (options.num_bookmarks > 0)
    num_records++;
  data_offset += num_records * sizeof(RecordEntryType);
  /* DWord align the actual data.  Don't know if this is needed, but
     DOC files are this way */
  if (data_offset % 4)
    data_offset += data_offset % 4;
  dbHeader->recordList.numRecords = htons(num_records);

  /* Initialize entries for record 0 */
  memset(&record0, 0, sizeof(zTXT_record0));
  record0.version = htons(options.version);
  /* numRecords should only count the number of compressed text records */
  if (options.num_bookmarks > 0)
    record0.numRecords = htons(num_records - 2);
  else
    record0.numRecords = htons(num_records - 1);
  record0.size = htonl(filesize);
  record0.recordSize = htons(RECORD_SIZE);
  record0.numBookmarks = htons(options.num_bookmarks);
  if (options.num_bookmarks > 0)
    record0.bookmarkRecord = htons(num_records - 1);
  else
    record0.bookmarkRecord = 0;
  record0.numAnnotations = 0;
  record0.annotationRecord = 0;

  /* Stick record 0 into the record list */
  dbRecordEntries->localChunkID = htonl(data_offset);
  dbRecordEntries->attributes = 0x40;
  x = htonl(0x00424200);
  y = (char *)&x;
  dbRecordEntries->uniqueID[0] = y[1];
  dbRecordEntries->uniqueID[1] = y[2];
  dbRecordEntries->uniqueID[2] = y[3];
  dbRecordEntries += 1;

  /* Fill out the database record entries */
  num_data_records = num_records;
  if (options.num_bookmarks > 0)
    num_data_records--;
  curr_record = 1;
  do {
    dbRecordEntries->localChunkID = htonl(data_offset + sizeof(zTXT_record0) +
                                          ((curr_record - 1) * RECORD_SIZE));
    dbRecordEntries->attributes = 0x40;
    x = htonl(0x00424200 + curr_record);
    y = (char *)&x;
    dbRecordEntries->uniqueID[0] = y[1];
    dbRecordEntries->uniqueID[1] = y[2];
    dbRecordEntries->uniqueID[2] = y[3];
    dbRecordEntries += 1;
    curr_record++;
  } while (curr_record < num_data_records);

  /* Add bookmark record.  This is done separately because the last record of
   * compressed text is probably not exactly recordSize bytes in length */
  if (options.num_bookmarks > 0)
    {
      dbRecordEntries->localChunkID = (data_offset +
                                       sizeof(zTXT_record0) +
                                       ((curr_record - 2) * RECORD_SIZE));
      x = zstream.total_out % RECORD_SIZE;
      if (x == 0)
        dbRecordEntries->localChunkID += RECORD_SIZE;
      else
        dbRecordEntries->localChunkID += x;
      dbRecordEntries->localChunkID = htonl(dbRecordEntries->localChunkID);
      dbRecordEntries->attributes = 0x40;
      x = htonl(0x00424200 + curr_record);
      y = (char *)&x;
      dbRecordEntries->uniqueID[0] = y[1];
      dbRecordEntries->uniqueID[1] = y[2];
      dbRecordEntries->uniqueID[2] = y[3];
    }




  /* Form DB filename */
  strncpy(outfilename, argv[last_arg], 128);
  /* First strip off any directory prefixes */
  y = strrchr(outfilename, '/');
  if (y)
    memmove(outfilename, y+1, strlen(y+1)+1);
  /* Now strip off any existing extension */
  y = strrchr(outfilename, '.');
  if (y == NULL)
    y = &outfilename[strlen(outfilename)];
  strcpy(y,".pdb");

  outfile = open(outfilename, O_WRONLY|O_CREAT|O_TRUNC,
                 S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
  errorif(outfile == -1);

  /* Output the DB header */
  x = write(outfile,header,data_offset);
  errorif(x != data_offset);

  /* Output record 0 */
  x = write(outfile, &record0, sizeof(zTXT_record0));
  errorif(x != sizeof(zTXT_record0));

  /* Output the compressed document */
  x = write(outfile,zbuf,zstream.total_out);
  errorif(x != zstream.total_out);

  /* Output the bookmarks */
  if (options.num_bookmarks > 0)
    {
      x = output_bookmarks(outfile);
      errorif(x != options.num_bookmarks * (MAX_BMRK_LENGTH + sizeof(UInt32)));
    }
  close(outfile);


  /* Clear the memory used by the bookmarks and regex */
  clear_regex_bmrks();

  free(header);
  free(inputbuf);
  free(zbuf);

  return 0;
}


/* Process commandline arguments */
int
process_arguments(int argc, char *argv[])
{
  int  ind;
  int  opt;
  char *usage = 
    "usage:   makeztxt [-hlV] [-t title] [-a adjust_type] [-b line_length]\n"
    "                  [-c creator] [-u type] [-d dbversion] [-w wbits]\n"
    "                  [-r regex] input\n";

  char *optstr = "a:b:c:d:r:t:u:w:hlV";
  static struct option long_options[] = {
    {"adjust", 1, 0, 'a'},
    {"length", 1, 0, 'b'},
    {"creator", 1, 0, 'c'},
    {"dbversion", 1, 0, 'd'},
    {"help", 0, 0, 'h'},
    {"list", 0, 0, 'l'},
    {"regex", 1, 0, 'r'},
    {"title", 1, 0, 't'},
    {"type", 1, 0, 'u'},
    {"version", 0, 0, 'V'},
    {"wbits", 1, 0, 'w'},
    {0, 0, 0, 0}};
 
  if (argc < 2)
    {
      fprintf(stderr, "%s\n", usage);
      fprintf(stderr, "use makeztxt -h for more detailed help\n");
      exit(1);
    }

  opt = getopt_long(argc, argv, optstr, long_options, &ind);
  while (opt != -1)
    {
      switch (opt)
        {
          case 'a':
            options.adjust_type = atoi(optarg);
            if ((options.adjust_type < 0) || (options.adjust_type > 2))
              {
                fprintf(stderr, "adjustment type out of range.  "
                        "Must be 0, 1, or 2\n");
                options.adjust_type = 0;
              }
            break;
          case 'b':
            options.line_length = atoi(optarg);
            break;
          case 'c':
            if (strlen(optarg) != 4)
              fprintf(stderr, "Creator must be a 4 character string\n");
            else
              options.creator = optarg;
            break;
          case 'd':
            options.version = (short)strtol(optarg, NULL, 0);
            break;
          case 'h':
            printf("%s\n\n", usage);
            printf(
              "\t-a, --adjust      method for linefeed stripping (0,1,2)\n"
              "\t\t0 = strip linefeeds for lines longer than length\n"
              "\t\t1 = strip linefeed if line has any text\n"
              "\t\t2 = no linefeed stripping\n"
              "\t-b, --length      lines longer than this get linefeed "
              "stripped\n"
              "\t                    (default = autodetect)\n"
              "\t-c, --creator     creator ID\n"
              "\t-d, --dbversion   database version\n"
              "\t-h, --help        this help\n"
              "\t-l, --list        list regex generated bookmarks\n"
              "\t-r, --regex       do a regex search for bookmarks\n"
              "\t-t, --title       title of database\n"
              "\t-u, --type        type ID\n"
              "\t-V, --version     makeztxt version\n"
              "\t-w, --wbits       controls zLib mem usage (10 - 15)\n\n");
            exit(1);
          case 'l':
            options.list_bmrks =  1;
            break;
          case 'r':
            add_regex(optarg);
            break;
          case 't':
            options.title = optarg;
            break;
          case 'u':
            if (strlen(optarg) != 4)
              fprintf(stderr, "Type must be a 4 character string\n");
            else
              options.type = optarg;
            break;
          case 'V':
            printf("makeztxt %s\n", VERSION);
            exit(0);
            break;
          case 'w':
            options.wbits = atoi(optarg);
            if ((options.wbits > 15) || (options.wbits < 10))
              {
                fprintf(stderr, "wbits out of range.  Must be 10 to 15\n");
                options.wbits = MAXWBITS;
              }
            break;
        }

      opt = getopt_long(argc, argv, optstr, long_options, &ind);
    }

  return optind;
}


/* Add a regular expression to the list */
void
add_regex(char *new_pattern)
{
  regex_node    *current = options.regex_list;

  if (current == NULL)
    {
      options.regex_list = current = (regex_node *)malloc(sizeof(regex_node));
      current->pattern = new_pattern;
      current->bad = 0;
      current->next = NULL;
    }
  else
    {
      while (current->next !=  NULL)
        current = current->next;
      current->next = (regex_node *)malloc(sizeof(regex_node));
      current = current->next;
      current->pattern = new_pattern;
      current->bad = 0;
      current->next = NULL;      
    }

  options.num_regex++;
}


/* Compile regex and then run them through the input buffer */
void
process_regex(char *input)
{
  regex_node    *current = options.regex_list;
  regmatch_t    matches[1];
  char          *err_string;
  int           str_len;
  int           i;
  int           err;
  u_long        start_pos;

  /* Compile the regular expressions */
  while (current != NULL)
    {
      current->prog = (regex_t *)malloc(8192);

      err = regcomp(current->prog, current->pattern,
                    REG_EXTENDED | REG_NEWLINE);
      if (err != 0)
        {
          str_len = regerror(err, current->prog, NULL, 0);
          err_string = (char *)malloc(str_len);
          regerror(err, current->prog, err_string, str_len);
          fprintf(stderr, "\nregular expression generated errors:\n"
                  "\t%s\n"
                  "\t%s\n", current->pattern, err_string);
          free(err_string);
          current->bad = 1;
        }
      current = current->next;
    }

  /* Run the regex over the input buffer */
  current = options.regex_list;
  for (i = 0; ((i < options.num_regex) && (current != NULL)); i++)
    {
      start_pos = 0;
      if (!current->bad)
        {
          while (regexec(current->prog,
                         input+start_pos, 1, matches, 0) != REG_NOMATCH)
            {
              add_bmrk(input, start_pos + matches[0].rm_so,
                       start_pos + matches[0].rm_eo);
              start_pos += matches[0].rm_eo;
            }
        }

      /* Don't need the compiled regex anymore */
      regfree(current->prog);
      free(current->prog);

      current = current->next;
    }
}


/* Adds a bookmark entry to the linked list in sorted order */
void
add_bmrk(char *input, u_long rm_so, u_long rm_eo)
{
  bmrk_node     *current = options.bookmarks;
  bmrk_node     *prev = NULL;
  char          title_buf[255];
  char          *pos;
  u_long        bytes;

  if (current == NULL)
    {
      options.bookmarks = current = (bmrk_node *)malloc(sizeof(bmrk_node));
      current->next = NULL;
    }
  else
    {
      while((current != NULL) && (current->offset < rm_so))
        {
          prev = current;
          current = current->next;
        }

      if (current == NULL)
        {
          /* Adding to the end of the list */
          prev->next = current = (bmrk_node *)malloc(sizeof(bmrk_node));
          current->next = NULL;
        }
      else if (prev == NULL)
        {
          /* Adding at the beginning of the list */
          current = (bmrk_node *)malloc(sizeof(bmrk_node));
          current->next = options.bookmarks;
          options.bookmarks = current;
        }
      else
        {
          /* Adding in the middle of the list */
          prev->next = (bmrk_node *)malloc(sizeof(bmrk_node));
          prev = prev->next;
          prev->next = current;
          current = prev;
        }
    }

  bytes = rm_eo - rm_so;
  if (bytes > 254)
    bytes = 254;
  strncpy(title_buf, &input[rm_so], bytes);
  title_buf[bytes] = '\0';

  pos = title_buf;
  while ((*pos == ' ') || (*pos == '\t'))
    pos++;
  if (*pos == '\0')
    {
      sprintf(title_buf, "Position %ld", rm_so);
      pos = title_buf;
    }
  strncpy(current->title, pos, MAX_BMRK_LENGTH);
  current->offset = rm_so;
  options.num_bookmarks++;
}


/* Unallocate all bookmark and regex nodes */
void
clear_regex_bmrks(void)
{
  regex_node    *regex = options.regex_list;
  regex_node    *regex_tmp;
  bmrk_node     *bmrk = options.bookmarks;
  bmrk_node     *bmrk_tmp;

  while (regex != NULL)
    {
      regex_tmp = regex;
      regex = regex->next;
      free(regex_tmp);
    }

  while (bmrk != NULL)
    {
      bmrk_tmp = bmrk;
      bmrk = bmrk->next;
      free(bmrk_tmp);
    }
}


/* Output the bookmarks to the output file */
u_int
output_bookmarks(int outfile)
{
  char          *buf;
  bmrk_node     *current = options.bookmarks;
  u_int         bytes = 0;

  buf = (char *)malloc(MAX_BMRK_LENGTH + sizeof(UInt32));

  while (current != NULL)
    {
      *((UInt32 *)buf) = htonl(current->offset);
      strncpy(&buf[sizeof(UInt32)], current->title, MAX_BMRK_LENGTH);
      bytes += write(outfile, buf, MAX_BMRK_LENGTH + sizeof(UInt32));
      current = current->next;
    }

  return bytes;
}


/*  Read in the input file and adjust the newlines for the Palm */
u_int
load_and_adjust(FILE *infile, char *buffer)
{
  long  total_length = 0;
  long  num_lines = 0;
  int   len;
  int   pos = 0;
  char  instr[255];

  if (options.line_length == 0)
    {
      /* If line_length is 0 we autodetect.  This involves computing the
       * average line length of the entire file and then subtracting 5 */
      while (fgets_nocr(instr, 255, infile))
        {
          num_lines++;
          total_length += strlen(instr);
        }
      rewind(infile);
      options.line_length = (total_length / num_lines) - 5;
  
      /* Probably shouldn't strip linefeeds of lines shorter than 20 characters
       * under any circumstances (unless explicity told to). */
      if (options.line_length < 20)
        options.line_length = 20;
    }

  buffer[0] = '\0';

  while (fgets_nocr(instr, 255, infile))
    {
      len = strlen(instr);

      switch (options.adjust_type)
        {
          case 0:
            /* A newline by itself is okay.  For normal text, strip the
               newline unless the line is less than options.line_length
               chars, then leave it alone. */
            if ((len > options.line_length) && (instr[len-1] == '\n'))
              instr[len-1] = ' ';
            break;

          case 1:
            /* A newline by itself is okay.  For other lines, remove all
               linefeeds */
            if ((len != 1) && (instr[len-1] == '\n'))
              instr[len-1] = ' ';
            break;

          case 2:
          default:
            /* Just leave the next alone.  No linefeed stripping. */
        }

      /* Append the input string to the input buffer */
      memcpy(&(buffer[pos]), instr, len);
      pos += len;
    }
  buffer[pos] = '\0';

  return pos;
}


/* Read a line from the file and remove CR */
char *
fgets_nocr(char *buffer, int bufsize, FILE *infile)
{
  int   len;
  char  *str;

  str = fgets(buffer, bufsize, infile);
  if (str != NULL)
    {
      /* got something, remove trailing CR character */
      len = strlen(str);

      if (len >= 2)
        {
          if (str[len - 2] == '\r')
            {
              str[len - 2] = '\n';
              str[len - 1] = 0;
            }
        }
    }

  return str;
}


/* Open a file in a users home dir or in the current dir */
FILE *
fopen_in_home_dir(char *file, char *mode)
{
  char *str = getenv("HOME");
  FILE *fd = NULL;
  char buff[512];

  if (str != NULL)
    {
      /* try in home directory */
      strcpy(buff, str);
      strcat(buff, "/");
      strcat(buff, file);

      fd = fopen(buff, mode);
    }

  if (fd == NULL)
    fd = fopen(file, mode);

  return fd;
}


/* Load the options file if it exists */
void
load_options(void)
{
  FILE  *optfd;
  char  buff[256];
  char  *ptr;
  int   len;

  optfd = fopen_in_home_dir(OPTIONS_FILE, "r");
  if (optfd != NULL)
    {
      buff[0] = 0;

      while (fgets_nocr(buff, 255, optfd) != NULL)
        {
          ptr = buff;

          if (!buff[0])
            break;

          while (*ptr == ' ' || *ptr == '\t' || *ptr == '\n')
            ++ptr;

          if (*ptr != '#' && *ptr)
            {
              /* not a comment line and not blank - process */
              len = strlen(ptr);

              if (ptr[len - 1] == '\n')
                ptr[len - 1] = 0;

              add_regex(strdup(ptr));
            }
          buff[0] = 0;
        }
      fclose(optfd);
    }
}


/*  Converts a 4 byte string Palm ID into a 4 byte integer */
u_long
id_to_int(char *idstr)
{
  if (strlen(idstr) != 4)
    return 0;

  return (0x1000000*(u_long)idstr[0] +
          0x10000*(u_long)idstr[1] +
          0x100*(u_long)idstr[2] +
          (u_long)idstr[3]);
}


/*  Print error message and exit if 'error' is true */
void
errorif(int error)
{
  if (error)
    {
      perror("makeztxt");
      exit(1);
    }
}
